@@ -1,3 +1,4 @@
+
 /**
  * RFX codec encoder
  *
@@ -40,14 +41,74 @@
 #include "rfxencode_diff_rlgr3.h"
 #include "rfxencode_rgb_to_yuv.h"
 
+#include "rfxencode_dwt_shift_rem.h"
+
 #ifdef RFX_USE_ACCEL_X86
 #include "x86/funcs_x86.h"
+#include "rfxencode_diff_count_sse2.h"
+#include "rfxencode_dwt_shift_rem_sse2.h"
 #endif
 
 #ifdef RFX_USE_ACCEL_AMD64
 #include "amd64/funcs_amd64.h"
+#include "rfxencode_diff_count_sse2.h"
+#include "rfxencode_dwt_shift_rem_sse2.h"
 #endif
 
+static void
+clear_encoder_rbs(struct rfxencode *enc)
+{
+    int index;
+    int jndex;
+    for (index = 0; index < RFX_MAX_RB_Y; ++index)
+    {
+        for (jndex = 0; jndex < RFX_MAX_RB_X; ++jndex)
+        {
+            free(enc->rbs[index][jndex]);
+            enc->rbs[index][jndex] = NULL;
+        }
+    }
+}
+
+static int
+rfxencode_reset_encoder(void *handle)
+{
+    struct rfxencode *enc;
+    enc = (struct rfxencode *) handle;
+
+    enc->frame_idx = 0;
+    enc->header_processed = 0;
+
+    memset(enc->a_buffer, 0, sizeof(enc->a_buffer));
+    memset(enc->y_r_buffer, 0, sizeof(enc->y_r_buffer));
+    memset(enc->u_g_buffer, 0, sizeof(enc->u_g_buffer));
+    memset(enc->v_b_buffer, 0, sizeof(enc->v_b_buffer));
+
+    memset(enc->dwt_buffer_a, 0, sizeof(enc->dwt_buffer_a));
+    enc->dwt_buffer = (sint16 *) (((size_t) (enc->dwt_buffer_a)) & ~15);
+
+    memset(enc->dwt_buffer1_a, 0, sizeof(enc->dwt_buffer1_a));
+    enc->dwt_buffer1 = (sint16 *) (((size_t) (enc->dwt_buffer1_a)) & ~15);
+
+    memset(enc->dwt_buffer2_a, 0, sizeof(enc->dwt_buffer2_a));
+    enc->dwt_buffer2 = (sint16 *) (((size_t) (enc->dwt_buffer2_a)) & ~15);
+
+    memset(enc->dwt_buffer3_a, 0, sizeof(enc->dwt_buffer3_a));
+    enc->dwt_buffer3 = (sint16 *) (((size_t) (enc->dwt_buffer3_a)) & ~15);
+
+    memset(enc->dwt_buffer4_a, 0, sizeof(enc->dwt_buffer4_a));
+    enc->dwt_buffer4 = (sint16 *) (((size_t) (enc->dwt_buffer4_a)) & ~15);
+
+    memset(enc->dwt_buffer5_a, 0, sizeof(enc->dwt_buffer5_a));
+    enc->dwt_buffer5 = (sint16 *) (((size_t) (enc->dwt_buffer5_a)) & ~15);
+
+    memset(enc->dwt_buffer6_a, 0, sizeof(enc->dwt_buffer6_a));
+    enc->dwt_buffer6 = (sint16 *) (((size_t) (enc->dwt_buffer6_a)) & ~15);
+
+    clear_encoder_rbs(enc);
+    return 0;
+}
+
 /******************************************************************************/
 int
 rfxcodec_encode_create_ex(int width, int height, int format, int flags,
@@ -68,6 +129,10 @@ rfxcodec_encode_create_ex(int width, int height, int format, int flags,
     enc->dwt_buffer = (sint16 *) (((size_t) (enc->dwt_buffer_a)) & ~15);
     enc->dwt_buffer1 = (sint16 *) (((size_t) (enc->dwt_buffer1_a)) & ~15);
     enc->dwt_buffer2 = (sint16 *) (((size_t) (enc->dwt_buffer2_a)) & ~15);
+    enc->dwt_buffer3 = (sint16 *) (((size_t) (enc->dwt_buffer3_a)) & ~15);
+    enc->dwt_buffer4 = (sint16 *) (((size_t) (enc->dwt_buffer4_a)) & ~15);
+    enc->dwt_buffer5 = (sint16 *) (((size_t) (enc->dwt_buffer5_a)) & ~15);
+    enc->dwt_buffer6 = (sint16 *) (((size_t) (enc->dwt_buffer6_a)) & ~15);
 
 #if defined(RFX_USE_ACCEL_X86)
     cpuid_x86(1, 0, &ax, &bx, &cx, &dx);
@@ -156,8 +221,24 @@ rfxcodec_encode_create_ex(int width, int height, int format, int flags,
     enc->format = format;
     enc->rfx_encode_rgb_to_yuv = rfx_encode_rgb_to_yuv;
     enc->rfx_encode_argb_to_yuva = rfx_encode_argb_to_yuva;
+    enc->rfx_encode_dwt_shift_rem = rfx_encode_dwt_shift_rem;
+    enc->rfx_encode_diff_count = rfx_encode_diff_count;
     /* assign encoding functions */
-    if (flags & RFX_FLAGS_NOACCEL)
+    if (flags & RFX_FLAGS_PRO1)
+    {
+        enc->pro_ver = 1;
+        if (flags & RFX_FLAGS_NOACCEL)
+        {
+        }
+        else if (enc->got_sse2)
+        {
+#if defined(RFX_USE_ACCEL_X86) || defined(RFX_USE_ACCEL_AMD64)
+            enc->rfx_encode_diff_count = rfx_encode_diff_count_sse2;
+            enc->rfx_encode_dwt_shift_rem = rfx_encode_dwt_shift_rem_sse2;
+#endif
+        }
+    }
+    else if (flags & RFX_FLAGS_NOACCEL)
     {
         if (enc->mode == RLGR3)
         {
@@ -301,6 +382,7 @@ rfxcodec_encode_destroy(void *handle)
     {
         return 0;
     }
+    clear_encoder_rbs(enc);
     free(enc);
     return 0;
 }
@@ -323,6 +405,32 @@ rfxcodec_encode_ex(void *handle, char *cdata, int *cdata_bytes,
     s.p = s.data;
     s.size = *cdata_bytes;
 
+    if (enc->pro_ver > 0)
+    {
+        if (flags & RFX_FLAGS_PRO_KEY)
+        {
+            rfxencode_reset_encoder(handle);
+        }
+        /* Only the first frame should send the RemoteFX header */
+        if ((enc->frame_idx == 0) && (enc->header_processed == 0))
+        {
+            if (rfx_pro_compose_message_header(enc, &s) != 0)
+            {
+                return -1;
+            }
+        }
+        tiles_written = rfx_pro_compose_message_data(enc, &s, regions, num_regions,
+                                         buf, width, height, stride_bytes,
+                                         tiles, num_tiles, quants, num_quants,
+                                         flags);
+        if (tiles_written <= 0)
+        {
+            return -1;
+        }
+        *cdata_bytes = (int) (s.p - s.data);
+        return tiles_written;
+    }
+
     /* Only the first frame should send the RemoteFX header */
     if ((enc->frame_idx == 0) && (enc->header_processed == 0))
     {
@@ -364,13 +472,64 @@ rfxcodec_encode_get_internals(struct rfxcodec_encode_internals *internals)
     internals->rfxencode_dwt_2d = rfx_dwt_2d_encode;
     internals->rfxencode_diff_rlgr1 = rfx_encode_diff_rlgr1;
     internals->rfxencode_diff_rlgr3 = rfx_encode_diff_rlgr3;
+    internals->rfx_encode_diff_count = rfx_encode_diff_count;
+    internals->rfx_encode_dwt_shift_rem = rfx_encode_dwt_shift_rem;
 #if defined(RFX_USE_ACCEL_X86)
     internals->rfxencode_dwt_shift_x86_sse2 = rfxcodec_encode_dwt_shift_x86_sse2;
     internals->rfxencode_dwt_shift_x86_sse41 = rfxcodec_encode_dwt_shift_x86_sse41;
+    internals->rfx_encode_diff_count_sse2 = rfx_encode_diff_count_sse2;
+    internals->rfx_encode_dwt_shift_rem_sse2 = rfx_encode_dwt_shift_rem_sse2;
 #endif
 #if defined(RFX_USE_ACCEL_AMD64)
     internals->rfxencode_dwt_shift_amd64_sse2 = rfxcodec_encode_dwt_shift_amd64_sse2;
     internals->rfxencode_dwt_shift_amd64_sse41 = rfxcodec_encode_dwt_shift_amd64_sse41;
+    internals->rfx_encode_diff_count_sse2 = rfx_encode_diff_count_sse2;
+    internals->rfx_encode_dwt_shift_rem_sse2 = rfx_encode_dwt_shift_rem_sse2;
 #endif
     return 0;
 }
+
+/*****************************************************************************/
+/* produce a hex dump */
+void
+rfxcodec_hexdump(const void *p, int len)
+{
+    unsigned char *line;
+    int i;
+    int thisline;
+    int offset;
+
+    line = (unsigned char *)p;
+    offset = 0;
+
+    while (offset < len)
+    {
+        printf("%04x ", offset);
+        thisline = len - offset;
+
+        if (thisline > 16)
+        {
+            thisline = 16;
+        }
+
+        for (i = 0; i < thisline; i++)
+        {
+            printf("%02x ", line[i]);
+        }
+
+        for (; i < 16; i++)
+        {
+            printf("   ");
+        }
+
+        for (i = 0; i < thisline; i++)
+        {
+            printf("%c", (line[i] >= 0x20 && line[i] < 0x7f) ? line[i] : '.');
+        }
+
+        printf("%s", "\n");
+        offset += thisline;
+        line += thisline;
+    }
+}
+
